@@ -1194,7 +1194,7 @@ def p_statement(s, level, cdef_flag = 0, visibility = 'private', api = 0):
             s.error("ctypedef statement not allowed here")
         if api:
             error(s.pos, "'api' not allowed with 'ctypedef'")
-        return p_ctypedef_statement(s, level, visibility)
+        return p_ctypedef_statement(s, level, visibility, api)
     elif s.sy == 'DEF':
         return p_DEF_statement(s)
     elif s.sy == 'IF':
@@ -1531,12 +1531,14 @@ def p_c_arg_list(s, in_pyfunc, cmethod_flag = 0, nonempty_declarators = 0,
     #  Comma-separated list of C argument declarations, possibly empty.
     #  May have a trailing comma.
     args = []
+    is_self_arg = cmethod_flag
     while s.sy not in c_arg_list_terminators:
-        args.append(p_c_arg_decl(s, in_pyfunc, cmethod_flag,
+        args.append(p_c_arg_decl(s, in_pyfunc, is_self_arg,
             nonempty = nonempty_declarators, kw_only = kw_only))
         if s.sy <> ',':
             break
         s.next()
+        is_self_arg = 0
     return args
 
 def p_optional_ellipsis(s):
@@ -1588,24 +1590,24 @@ def p_cdef_statement(s, level, visibility = 'private', api = 0):
     if visibility == 'extern' and s.sy == 'from':
             return p_cdef_extern_block(s, level, pos)
     elif s.sy == ':':
-        p_cdef_block(s, level, visibility, api)
+        return p_cdef_block(s, level, visibility, api)
     elif s.sy == 'class':
         if level not in ('module', 'module_pxd'):
             error(pos, "Extension type definition not allowed here")
-        if api:
-            error(pos, "'api' not allowed with extension class")
-        return p_c_class_definition(s, level, pos, visibility = visibility)
+        #if api:
+        #	error(pos, "'api' not allowed with extension class")
+        return p_c_class_definition(s, level, pos, visibility = visibility, api = api)
     elif s.sy == 'IDENT' and s.systring in struct_union_or_enum:
         if level not in ('module', 'module_pxd'):
             error(pos, "C struct/union/enum definition not allowed here")
         #if visibility == 'public':
         #	error(pos, "Public struct/union/enum definition not implemented")
-        if api:
-            error(pos, "'api' not allowed with '%s'" % s.systring)
+        #if api:
+        #	error(pos, "'api' not allowed with '%s'" % s.systring)
         if s.systring == "enum":
-            return p_c_enum_definition(s, pos, visibility)
+            return p_c_enum_definition(s, pos, level, visibility)
         else:
-            return p_c_struct_or_union_definition(s, pos, visibility)
+            return p_c_struct_or_union_definition(s, pos, level, visibility)
     elif s.sy == 'pass':
         node = p_pass_statement(s)
         s.expect_newline('Expected a newline')
@@ -1614,8 +1616,7 @@ def p_cdef_statement(s, level, visibility = 'private', api = 0):
         return p_c_func_or_var_declaration(s, level, pos, visibility, api)
 
 def p_cdef_block(s, level, visibility, api):
-    body = p_suite(s, level, cdef_flag = 1, visibility = 'extern', api = api)
-    return Nodes.StatListNode(pos, stats = body)
+    return p_suite(s, level, cdef_flag = 1, visibility = visibility, api = api)
 
 def p_cdef_extern_block(s, level, pos):
     include_file = None
@@ -1633,7 +1634,7 @@ struct_union_or_enum = (
     "struct", "union", "enum"
 )
 
-def p_c_enum_definition(s, pos, visibility, typedef_flag = 0):
+def p_c_enum_definition(s, pos, level, visibility, typedef_flag = 0):
     # s.sy == ident 'enum'
     s.next()
     if s.sy == 'IDENT':
@@ -1656,7 +1657,8 @@ def p_c_enum_definition(s, pos, visibility, typedef_flag = 0):
             p_c_enum_line(s, items)
         s.expect_dedent()
     return Nodes.CEnumDefNode(pos, name = name, cname = cname,
-        items = items, typedef_flag = typedef_flag, visibility = visibility)
+        items = items, typedef_flag = typedef_flag, visibility = visibility,
+        in_pxd = level == 'module_pxd')
 
 def p_c_enum_line(s, items):
     if s.sy <> 'pass':
@@ -1681,7 +1683,7 @@ def p_c_enum_item(s, items):
     items.append(Nodes.CEnumDefItemNode(pos, 
         name = name, cname = cname, value = value))
 
-def p_c_struct_or_union_definition(s, pos, visibility, typedef_flag = 0):
+def p_c_struct_or_union_definition(s, pos, level, visibility, typedef_flag = 0):
     # s.sy == ident 'struct' or 'union'
     kind = s.systring
     s.next()
@@ -1706,7 +1708,8 @@ def p_c_struct_or_union_definition(s, pos, visibility, typedef_flag = 0):
         s.expect_newline("Syntax error in struct or union definition")
     return Nodes.CStructOrUnionDefNode(pos, 
         name = name, cname = cname, kind = kind, attributes = attributes,
-        typedef_flag = typedef_flag, visibility = visibility)
+        typedef_flag = typedef_flag, visibility = visibility,
+        in_pxd = level == 'module_pxd')
 
 def p_visibility(s, prev_visibility):
     pos = s.position()
@@ -1734,8 +1737,8 @@ def p_c_func_or_var_declaration(s, level, pos, visibility = 'private', api = 0):
             body = suite,
             api = api)
     else:
-        if api:
-            error(s.pos, "'api' not allowed with variable declaration")
+        #if api:
+        #	error(pos, "'api' not allowed with variable declaration")
         declarators = [declarator]
         while s.sy == ',':
             s.next()
@@ -1748,29 +1751,31 @@ def p_c_func_or_var_declaration(s, level, pos, visibility = 'private', api = 0):
             visibility = visibility,
             base_type = base_type, 
             declarators = declarators,
-            in_pxd = level == 'module_pxd')
+            in_pxd = level == 'module_pxd',
+            api = api)
     return result
 
-def p_ctypedef_statement(s, level, visibility = 'private'):
+def p_ctypedef_statement(s, level, visibility = 'private', api = 0):
     # s.sy == 'ctypedef'
     pos = s.position()
     s.next()
     visibility = p_visibility(s, visibility)
     if s.sy == 'class':
         return p_c_class_definition(s, level, pos,
-            visibility = visibility,
-            typedef_flag = 1)
+            visibility = visibility, typedef_flag = 1, api = api)
     elif s.sy == 'IDENT' and s.systring in ('struct', 'union', 'enum'):
         if s.systring == 'enum':
-            return p_c_enum_definition(s, pos, visibility, typedef_flag = 1)
+            return p_c_enum_definition(s, pos, level, visibility, typedef_flag = 1)
         else:
-            return p_c_struct_or_union_definition(s, pos, visibility, typedef_flag = 1)
+            return p_c_struct_or_union_definition(s, pos, level, visibility,
+                typedef_flag = 1)
     else:
         base_type = p_c_base_type(s)
         declarator = p_c_declarator(s, is_type = 1, nonempty = 1)
         s.expect_newline("Syntax error in ctypedef statement")
         return Nodes.CTypeDefNode(pos,
-            base_type = base_type, declarator = declarator, visibility = visibility)
+            base_type = base_type, declarator = declarator, visibility = visibility,
+            in_pxd = level == 'module_pxd')
 
 def p_def_statement(s):
     # s.sy == 'def'
@@ -1826,7 +1831,7 @@ def p_class_statement(s):
         doc = doc, body = body)
 
 def p_c_class_definition(s, level, pos, 
-        visibility = 'private', typedef_flag = 0):
+        visibility = 'private', typedef_flag = 0, api = 0):
     # s.sy == 'class'
     s.next()
     module_path = []
@@ -1882,9 +1887,13 @@ def p_c_class_definition(s, level, pos,
             error(pos, "Object struct name specification required for 'public' C class")
         if not typeobj_name:
             error(pos, "Type object name specification required for 'public' C class")
+    else:
+        if api:
+            error(pos, "Only 'public' C class can be declared 'api'")
     return Nodes.CClassDefNode(pos,
         visibility = visibility,
         typedef_flag = typedef_flag,
+        api = api,
         module_name = ".".join(module_path),
         class_name = class_name,
         as_name = as_name,
